# -*- coding: utf-8 -*-

#  PycaWM
#  client.py
#  Copyright (c) 2007 Vincent Rasneur, Anaël Verrier

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; version 3 only.

#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import sys

if sys.version_info < (2, 5):
    raise ImportError('client module needs at least Python 2.5!')

import Image

from Xlib import X, Xutil, protocol, error
from Xlib.error import BadDrawable, BadWindow, BadMatch
from Xlib.X import ZPixmap
from Xlib.ext import shape

from atomsdb import atomsdb
from hookmanager import MetaHookable, disable_method_for_instance
from misc import good_key, good_grab_key, good_ungrab_key, numlock_mask
from patterns import Null, DictWrapper


class PseudoParent(object):
    overrided_methods = ['map', 'unmap', 'destroy']

    def __init__(self, client):
        self.client = client

    def __eq__(self, window):
        return self.client.window == window

    def do_nothing(self):
        pass

    def __getattr__(self, attr):
        if attr in self.overrided_methods:
            return self.do_nothing
        else:
            # delegate the methods to the real window
            return getattr(self.client.window, attr)

class Client(object):
    __metaclass__ = MetaHookable
    parent_border_width = 2
    title_bar_height = 16
    # clients list ordered by focus
    clients = list()
    stem_buttons = [[], []]
    keys = dict()
    left, right = range(2)
    # maximization state
    (unmaximized, maximized_vert, maximized_hort, maximized_both,
     taken_all_vert_space, taken_all_hort_space, taken_all_free_space,
     fullscreened) = range(8)
    # client "ICCCM" state (in fact ICCCM + "just_reparented" state)
    withdrawn, normal, just_reparented, iconified = range(4)
    # initial client states (each state can be simutaneously active)
    initial_states = dict(
        icccm_state=withdrawn,
        reparented=False,
        focused=False,
        grabbed=False,
        transient=False,
        shaded=False,
        maximized=unmaximized)
    # local error handler for all the clients
    error_handler = error.CatchError(BadWindow, BadMatch)

    def __init__(self, window, parent, wm):
        self.wm = wm
        self.display = wm.display
        self.window = window
        self.parent = parent
        self.title_bar = None
        self.screen = wm.screen
        self.window.change_save_set(X.SetModeInsert)
        self.window.set_wm_state(state=Xutil.WithdrawnState, icon=0)
        self.states = DictWrapper(self.initial_states.copy())
        self.states.maximized = self.unmaximized
        self.states.icccm_state = self.withdrawn
        self.transient_for = self.window.get_wm_transient_for()
        self.states.transient = self.transient_for is not None
        # if the client window is transient,
        # add it to the main client transient clients list
        if self.states.transient:
            client = self.wm.get_managed_client(self.transient_for)
            if client is not None and self not in client.transient_clients:
                client.transient_clients.append(self)
        # build the transient clients list for the client we are adding
        self.transient_clients = list()
        for client in self.clients:
            if self.window == client.transient_for:
                self.transient_clients.append(client)
        self.name = self.window.get_wm_name()
        self.geometry = self.window.get_geometry()
        self.wm_normal_hints = self.window.get_wm_normal_hints()
        self.wm_hints = self.window.get_wm_hints()
        self.motif_wm_hints = self.window.get_property(atomsdb._MOTIF_WM_HINTS,
                                                       atomsdb._MOTIF_WM_HINTS,
                                                       0, 5)
        self.protocols = self.window.get_wm_protocols()
        self.gc = None
        for keysym, mask in self.keys:
            good_grab_key(self.window, self.display.keysym_to_keycode(keysym),
                          mask)
        self.clients.append(self)
        self.coords_backup = None
        self.buttons = [[], []]

    def make_first(self):
        try:
            self.clients.remove(self)
            self.clients.insert(0, self)
        except AttributeError:
            print '%s is not on the client list?!' % self

    def make_last(self):
        try:
            self.clients.remove(self)
            self.clients.append(self)
        except AttributeError:
            print '%s is not on the client list?!' % self

    def is_valid(self):
        # is the client waiting to be destroyed?
        # if so, the window does not exist anymore
        self.wm.display.flush()
        return self.window in self.parent.query_tree().children

    def is_hidden(self):
        return (self.states.icccm_state == self.withdrawn or
                self.states.icccm_state == self.iconified)

    def grab_button(self):
        if self.states.grabbed:
            return

        self.window.grab_button(X.AnyButton, X.AnyModifier,
                                False, X.ButtonPressMask,
                                X.GrabModeSync, X.GrabModeSync,
                                X.NONE, X.NONE)
        self.states.grabbed = True

    def ungrab_button(self):
        if not self.states.grabbed:
            return

        self.window.ungrab_button(X.AnyButton, X.AnyModifier)
        self.states.grabbed = False

    @classmethod
    def add_button(cls, button, where):
        cls.stem_buttons[where].append(button)
        for client in cls.clients:
            client.buttons[where].append(button.copy(client))
            client.update_frame()

    @classmethod
    def add_left_button(cls, button):
        cls.add_button(button, cls.left)

    @classmethod
    def add_right_button(cls, button):
        cls.add_button(button, cls.right)

    @classmethod
    def remove_button(cls, button, where):
        # to not have a circular import...
        from button import Button
        if isinstance(button, Button):
            if button in cls.stem_buttons[where]:
                indice = cls.stem_buttons[where].index(button)
            else:
                return
        else:
            try:
                indice = int(button)
            except ValueError:
                return

        if indice >= len(cls.stem_buttons[where]):
            return

        del cls.stem_buttons[where][indice]
        for client in cls.clients:
            del client.buttons[where][indice]
            client.update_frame()

    @classmethod
    def remove_left_button(cls, button):
        cls.remove_button(button, cls.left)

    @classmethod
    def remove_right_button(cls, button):
        cls.remove_button(button, cls.right)

    def button_pressed(self, x, y):
        button_x = 0
        for button in self.buttons[self.left]:
            if not button.client_p(self):
                continue
            if (x >= button_x and x < button_x + button.size[0] and
                y >= 0 and y < button.size[1]):
                return button
            button_x += button.size[0]
        button_x = self.geometry.width
        for button in self.buttons[self.right]:
            if not button.client_p(self):
                continue
            if (x >= button_x - button.size[0] and x < button_x and
                y >= 0 and y < button.size[1]):
                return button
            button_x -= button.size[0]
        return None

    def button_press(self, x, y, mouse_button):
        button = self.button_pressed(x, y)
        if button is not None:
            button.press(self, mouse_button)
        return button

    @classmethod
    def add_key(cls, keys, fun):
        keysym, mask = good_key(keys)
        if keysym != X.NoSymbol and mask != 0:
            if (keysym, mask) not in cls.keys:
                cls.keys[keysym, mask] = fun
                if cls.clients:
                    # update the key grabbing for all the clients
                    display = atomsdb.display
                    keysym_to_keycode = display.keysym_to_keycode
                    for client in cls.clients:
                        good_grab_key(client.window,
                                      keysym_to_keycode(keysym), mask)
                    display.sync()

    @classmethod
    def remove_key(cls, keys):
        keysym, mask = good_key(keys)
        if keysym != X.NoSymbol and mask != 0:
            if (keysym, mask) in cls.keys:
                del cls.keys[keysym, mask]
                if cls.clients:
                    # update the key grabbing for all the clients
                    display = atomsdb.display
                    keysym_to_keycode = display.keysym_to_keycode
                    for client in cls.clients:
                        good_ungrab_key(client.window,
                                        keysym_to_keycode(keysym), mask)
                    display.sync()

    def show_buttons(self, pixmap):
        button_x = 0
        for button in self.buttons[self.left]:
            if button.client_p(self):
                button.show(self, button_x, 0, pixmap)
                button_x += button.size[0]
        button_x = self.geometry.width
        for button in self.buttons[self.right]:
            if button.client_p(self):
                button_x -= button.size[0]
                button.show(self, button_x, 0, pixmap)

    def update_frame(self):
        if self.gc is not None and self.states.reparented:
            title_bar_pixmap = self.title_bar.create_pixmap(
                self.geometry.width,
                self.title_bar_height,
                self.screen.root_depth)
            # clear the pixmap
            self.gc.change(foreground=self.wm.get_color(
                    'black' if self.states.focused else 'DimGray'))
            title_bar_pixmap.fill_rectangle(self.gc, 0, 0,
                                            self.geometry.width,
                                            self.title_bar_height)
            # window name
            text_x = 5
            for button in self.buttons[self.left]:
                if button.client_p(self):
                    text_x += button.size[0]
            self.gc.change(foreground=self.wm.get_color('white'),
                           background=self.wm.get_color(
                    'black' if self.states.focused else 'DimGray'))
            title_bar_pixmap.image_text(self.gc, text_x, 10, str(self.name))
            # buttons
            self.show_buttons(title_bar_pixmap)
            self.title_bar.copy_area(self.gc, title_bar_pixmap, 0, 0,
                                     self.geometry.width,
                                     self.title_bar_height,
                                     0, 0)
            title_bar_pixmap.free()

    def map(self):
        if (self.states.icccm_state == self.just_reparented or
            self.states.icccm_state == self.normal):
            # client getting a map after the reparenting
            # or mapped client getting a map again
            self.parent.map()
            self.title_bar.map()
            self.window.map()
        # unmapped client getting a map...
        elif (self.parent != self.screen.root and
              self.states.icccm_state != self.normal):
            self.window.change_save_set(X.SetModeInsert)
            self.unhide()
            self.states.icccm_state = self.normal

        self.window.set_wm_state(state=Xutil.NormalState, icon=0)
        self.update_frame()

    def fake_reparent(self):
        # fake reparenting already done?
        if isinstance(self.parent, PseudoParent):
            return

        self.parent = PseudoParent(self)
        self.title_bar = Null()
        self.title_bar_height = 0
        self.parent_border_width = 0
        self.window.change_attributes(event_mask=(X.ColormapChangeMask |
                                                  X.PropertyChangeMask |
                                                  X.EnterWindowMask |
                                                  X.LeaveWindowMask),
                                      win_gravity=X.StaticGravity)
        self.states.icccm_state = self.normal
        self.states.reparented = False

    def reparent(self):
        # do not reparent an already reparented window!
        if self.parent != self.screen.root:
            return

        # take care of xmms style windows
        if (self.motif_wm_hints is not None and
            self.motif_wm_hints.value[0] == 2):
            self.fake_reparent()
            return

        # create the window frame
        self.parent = self.screen.root.create_window(
            self.geometry.x,
            self.geometry.y,
            self.geometry.width,
            self.geometry.height + self.title_bar_height,
            1, self.screen.root_depth,
            background_pixel=self.screen.black_pixel)
        self.parent.configure(border_width=self.parent_border_width)

        self.title_bar = self.screen.root.create_window(0, 0,
                                                        self.geometry.width,
                                                        self.title_bar_height,
                                                        1,
                                                        self.screen.root_depth)
        self.title_bar.configure(border_width=0)
        self.title_bar.change_attributes(event_mask=(X.ExposureMask |
                                                     X.KeyPressMask |
                                                     X.EnterWindowMask |
                                                     X.LeaveWindowMask |
                                                     X.ButtonPressMask |
                                                     X.ButtonReleaseMask))
        self.gc = self.title_bar.create_gc(foreground=self.screen.white_pixel)

        self.parent.change_attributes(event_mask=(X.ExposureMask |
                                                  X.KeyPressMask |
                                                  X.EnterWindowMask |
                                                  X.LeaveWindowMask |
                                                  X.ButtonPressMask |
                                                  X.ButtonReleaseMask |
                                                  X.SubstructureRedirectMask |
                                                  X.SubstructureNotifyMask))
        self.window.change_attributes(event_mask=(X.ColormapChangeMask |
                                                  X.PropertyChangeMask |
                                                  X.EnterWindowMask),
                                      do_not_propagate_mask=X.ButtonPressMask |
                                      X.ButtonReleaseMask,
                                      win_gravity=X.StaticGravity)
        self.window.configure(border_width=0)
        self.window.reparent(self.parent, 0, self.title_bar_height)
        self.title_bar.reparent(self.parent, 0, 0)
        # create the buttons from the stem buttons
        for side in (self.left, self.right):
            for stem_button in self.stem_buttons[side]:
                self.buttons[side].append(stem_button.copy(self))
        self.geometry = self.parent.get_geometry()
        self.states.icccm_state = self.just_reparented
        self.states.reparented = True
        if self.display.has_extension('SHAPE'):
            self.configure_shape()
        self.display.sync()

    def withdraw(self):
        if self.states.icccm_state == self.withdrawn:
            return

        # synthetic UnmapNotify from the reparenting?
        if self.states.icccm_state == self.just_reparented:
            self.states.icccm_state = self.normal
            return

        self.states.icccm_state = self.withdrawn
        # the 2 lines below will generate BadWindow errors
        # if the client window is destroyed
        self.window.set_wm_state(state=Xutil.WithdrawnState, icon=0,
                                 onerror=self.error_handler)
        self.window.change_save_set(X.SetModeDelete, onerror=self.error_handler)

        # iconified window disappearing?
        if self.states.icccm_state == self.iconified:
            # do nothing...
            pass
        # normal window?
        elif self.parent != self.screen.root:
            self.parent.unmap()
            self.title_bar.unmap()
            self.states.focused = False
            self.make_last()
            self.wm.set_new_focused_client()

    def manage(self):
        self.reparent()
        if (self.wm_hints is not None and
            self.wm_hints.initial_state == Xutil.IconicState):
            self.iconify()
        else:
            self.wm.place_client(self)
            self.map()

    def unmanage(self):
        self.window.ungrab_key(X.AnyKey, X.AnyModifier)
        self.window.reparent(self.screen.root, self.geometry.x, self.geometry.y)
        self.title_bar.reparent(self.screen.root, 0, 0)
        self.title_bar.destroy()
        self.parent.destroy()
        self.parent = self.screen.root
        self.window.change_attributes(event_mask=(X.ExposureMask |
                                                  X.EnterWindowMask |
                                                  X.ButtonPressMask |
                                                  X.ButtonReleaseMask |
                                                  X.SubstructureRedirectMask |
                                                  X.SubstructureNotifyMask))
        self.window.configure(stack_mode = X.Below)
        self.clients.remove(self)

    def contain(self, point_x, point_y):
        geometry = self.geometry
        return (point_x >= geometry.x and
                point_x < (geometry.x + geometry.width +
                           (self.parent_border_width * 2)) and
                point_y >= geometry.y and
                point_y < (geometry.y + geometry.height +
                           (self.parent_border_width * 2)))

    def window_contain(self, point_x, point_y):
        geometry = self.geometry
        return (
            point_x >= geometry.x + self.parent_border_width and
            point_x < geometry.x + geometry.width + self.parent_border_width and
            point_y >= geometry.y + self.parent_border_width + \
                self.title_bar_height and
            point_y < geometry.y + geometry.height + self.parent_border_width)

    def collide(self, client):
        geo = self.geometry
        geo_client = client.geometry
        self_border = self.parent_border_width
        client_border = client.parent_border_width
        return (geo.x < geo_client.x + geo_client.width + 2 * client_border and
                geo.x + geo.width + 2 * self_border > geo_client.x and
                geo.y < geo_client.y + geo_client.height + 2 * client_border and
                geo.y + geo.height + 2 * self_border > geo_client.y)

    def update_geometry(self, x=None, y=None, width=None, height=None):
        if x is not None:
            self.geometry.x = x
        if y is not None:
            self.geometry.y = y
        if width is not None:
            self.geometry.width = width
        if height is not None:
            self.geometry.height = height

    def move(self, coords):
        self.parent.configure(**coords)
        self.geometry = self.parent.get_geometry()

    def update_size(self, change):
        for attr in change:
            if attr == 'x':
                self.parent.configure(x=change[attr])
            elif attr == 'y':
                self.parent.configure(y=change[attr])
            elif attr == 'width':
                self.parent.configure(width=change[attr])
                self.title_bar.configure(width=change[attr])
                self.window.configure(width=change[attr])
            elif attr == 'height':
                self.parent.configure(height=change[attr] +
                                      self.title_bar_height)
                self.window.configure(height=change[attr])
            elif attr == 'stack_mode':
                if change[attr] == X.Above:
                    self.upper()
                else: # change[attr] == X.Below
                    self.lower()

        self.geometry = self.parent.get_geometry()

    def good_dx_increment(self, dx, little=False):
        if (self.wm_normal_hints is not None and
            self.wm_normal_hints.flags & Xutil.PResizeInc):
            if (dx != 0 and dx % self.wm_normal_hints.width_inc):
                dx -= dx % self.wm_normal_hints.width_inc
                # (dx > 0 and not little) or (dx < 0 and little)
                # is the same as (dx < 0) <=> little
                # so we write (dx < 0) == little in python
                if (dx < 0) == little:
                    dx += self.wm_normal_hints.width_inc
        return dx

    def good_dy_increment(self, dy, little=False):
        if (self.wm_normal_hints is not None and
            self.wm_normal_hints.flags & Xutil.PResizeInc):
            if (dy != 0 and dy % self.wm_normal_hints.height_inc):
                dy -= dy % self.wm_normal_hints.height_inc
                # (dy > 0 and not little) or (dy < 0 and little)
                # is the same as (dy < 0) <=> little
                # so we write (dy < 0) == little in python
                if (dy < 0) == little:
                    dy += self.wm_normal_hints.height_inc
        return dy

    def resize(self, coords, little=False):
        if self.states.shaded:
            # for now, do not resize a shaded client
            return

        left = coords['left']
        top = coords['top']
        right = coords['right']
        bottom = coords['bottom']

        if left == 0 and top == 0 and right == 0 and bottom == 0:
            return

        left = self.good_dx_increment(left, little)
        top = self.good_dy_increment(top, little)
        right = self.good_dx_increment(right, little)
        bottom = self.good_dy_increment(bottom, little)

        x = self.geometry.x + left
        y = self.geometry.y + top
        width = self.geometry.width - left + right
        height = self.geometry.height - self.title_bar_height - top + bottom

        if self.wm_normal_hints is not None:
            if (self.wm_normal_hints.flags & Xutil.PMinSize and
                (width < self.wm_normal_hints.min_width or
                 height < self.wm_normal_hints.min_height)):
                return
            if (self.wm_normal_hints.flags & Xutil.PMaxSize and
                (width > self.wm_normal_hints.max_width or
                 height > self.wm_normal_hints.max_height)):
                return
        else:
            if width < 5 or height < 5:
                return

        self.title_bar.configure(width=width, height=self.title_bar_height)
        self.parent.configure(width=width, height=height+self.title_bar_height)
        self.window.configure(width=width, height=height)
        self.move(dict(x=x, y=y))

    def upper(self):
        client = self.wm.get_managed_client(self.transient_for)
        if client is None:
            client = self
        # raise the client
        client.parent.configure(stack_mode=X.Above)
        # ... and all its transient windows
        for transient in client.transient_clients:
            transient.parent.configure(stack_mode=X.Above)

    def lower(self):
        unmanaged_windows = list()
        for window in self.screen.root.query_tree().children:
            managed = self.wm.get_managed_client(window)
            if managed is None:
                unmanaged_windows.append(window)
            elif managed == self:
                break

        client = self.wm.get_managed_client(self.transient_for)
        if client is None:
            client = self
        # lower the client's transient windows
        for transient in client.transient_clients:
            transient.parent.configure(stack_mode=X.Below)
        # ... and lower the client itself
        client.parent.configure(stack_mode=X.Below)
        # do not forget the unmanaged windows
        for window in reversed(unmanaged_windows):
            window.configure(stack_mode=X.Below)

    def leave_focus(self):
        self.states.focused = False
        self.update_frame()

    def take_focus(self):
        if self.states.focused or self.is_hidden():
            return

        if self.clients[0] != self:
            self.make_first()

        self.states.focused = True
        self.update_frame()
        if len(self.clients) > 1:
            self.clients[1].leave_focus()
        # the wm will sometimes give the focus to already destroyed windows
        # we will use the Client local error handler to catch
        # the generated errors
        if self.wm_hints is not None and not self.wm_hints.input:
            # cannot make the window focused
            # but the focus needs to be here, so focus the parent!
            self.parent.set_input_focus(X.RevertToPointerRoot, X.CurrentTime,
                                        onerror=self.error_handler)
        else:
            self.window.set_input_focus(X.RevertToPointerRoot, X.CurrentTime,
                                        onerror=self.error_handler)
            if atomsdb.WM_TAKE_FOCUS in self.protocols:
                focus_event = protocol.event.ClientMessage(
                    window=self.window,
                    client_type=atomsdb.WM_PROTOCOLS,
                    data=(32, [atomsdb.WM_TAKE_FOCUS, X.CurrentTime, 0, 0, 0]))
                self.window.send_event(focus_event, onerror=self.error_handler)

    def close(self):
        if atomsdb.WM_DELETE_WINDOW in self.protocols:
            delete_event = protocol.event.ClientMessage(
                window=self.window,
                client_type=atomsdb.WM_PROTOCOLS,
                data=(32, [atomsdb.WM_DELETE_WINDOW, X.CurrentTime, 0, 0, 0]))
            self.window.send_event(delete_event)
        else:
            self.window.kill_client()

    def destroy(self):
        if self.title_bar is not None:
            self.title_bar.reparent(self.screen.root, 0, 0)
            self.title_bar.destroy()
        if self.parent != self.screen.root:
            self.parent.destroy()
        if self.gc is not None:
            self.gc.free()
        client = self.wm.get_managed_client(self.transient_for)
        if client is not None and self in self.transient_clients:
            client.transient_clients.remove(self)
        del self.transient_clients[:]
        self.clients.remove(self)

    def maximize(self, horizontally=True, vertically=True):
        if self.states.maximized:
            return
        if horizontally and vertically:
            self.states.maximized = self.maximized_both
        geo = self.geometry
        self.coords_backup = (geo.x, geo.y,
                              geo.width, geo.height)
        if horizontally:
            x = 0
            right = self.screen.width_in_pixels - (1 + geo.width)
        else:
            x = geo.x
            right = 0
        if vertically:
            y = 0
            bottom = self.screen.height_in_pixels - (1 + geo.height)
        else:
            y = geo.y
            bottom = 0
        self.move(dict(x=x, y=y))
        self.resize(dict(left=0, top=0, right=right, bottom=bottom), True)

    def maximize_horizontally(self):
        if self.states.maximized:
            return
        self.maximize(vertically=False)
        self.states.maximized = self.maximized_hort

    def maximize_vertically(self):
        if self.states.maximized:
            return
        self.maximize(horizontally=False)
        self.states.maximized = self.maximized_vert

    def unmaximize(self):
        if not self.states.maximized:
            return
        if self.states.maximized == self.fullscreened:
            # Warning: alpha code ;-)
            self.parent.configure(border_width=self.parent_border_width)
            self.window.configure(x=0, y=self.title_bar_height)
            self.title_bar.map()
        geo = self.geometry
        x, y = self.coords_backup[:2]
        right = self.coords_backup[2] - geo.width
        bottom = self.coords_backup[3] - geo.height
        self.resize(dict(left=0, top=0, right=right, bottom=bottom), True)
        self.move(dict(x=x, y=y))
        self.states.maximized = self.unmaximized

    def take_all_free_space(self, horizontally=True, vertically=True):
        if self.states.maximized:
            return
        if horizontally and vertically:
            self.states.maximized = self.taken_all_free_space
        clients = list()
        for client in self.clients:
            if not (client.is_hidden() or self.collide(client)):
                clients.append(client)
        geo = self.geometry
        self.coords_backup = (geo.x, geo.y,
                              geo.width, geo.height)
        if horizontally:
            x1, x2 = [-1], [self.screen.width_in_pixels]
        else:
            x1, x2 = [geo.x - 1], [geo.x + geo.width + 1]
        if vertically:
            y1, y2 = [-1], [self.screen.height_in_pixels]
        else:
            y1, y2 = [geo.y - 1], [geo.y + geo.height + 1]
        for client in clients:
            geo_client = client.geometry
            if (horizontally and
                geo_client.y < geo.y + geo.height and
                geo_client.y + geo_client.height > geo.y):
                if geo_client.x + geo_client.width < geo.x:
                    x1.append(geo_client.x + geo_client.width + 1)
                else:
                    x2.append(geo_client.x - 1)
            if (vertically and
                geo_client.x < geo.x + geo.width and
                geo_client.x + geo_client.width > geo.x):
                if geo_client.y + geo_client.height < geo.y:
                    y1.append(geo_client.y + geo_client.height + 1)
                else:
                    y2.append(geo_client.y - 1)
        self.move(dict(x=max(x1) + 1, y=max(y1) + 1))
        self.resize(
            dict(left=0,
                 top=0,
                 right=min(x2) - (max(x1) + geo.width),
                 bottom=min(y2) - (max(y1) + geo.height)),
            True)

    def take_all_horizontal_space(self):
        if self.states.maximized:
            return
        self.take_all_free_space(vertically=False)
        self.states.maximized = self.taken_all_hort_space

    def take_all_vertical_space(self):
        if self.states.maximized:
            return
        self.take_all_free_space(horizontally=False)
        self.states.maximized = self.taken_all_vert_space

    def fullscreen(self):
        # Warning: alpha code ;-)
        if self.states.maximized:
            return
        geo = self.geometry
        self.maximize()
        self.title_bar.unmap()
        self.parent.configure(border_width=0)
        self.window.configure(x=0, y=0,
                              width=geo.width + self.good_dx_increment(
            self.screen.width_in_pixels - geo.width, True),
                              height=geo.height + self.good_dy_increment(
            self.screen.height_in_pixels - geo.height, True))
        self.states.maximized = self.fullscreened

    def hide(self):
        # do not generate dummy enternotify events, unmap the parent first
        self.parent.unmap()
        self.title_bar.unmap()
        disable_method_for_instance(self.withdraw)
        self.window.unmap()
        self.states.focused = False
        self.make_last()
        self.wm.set_new_focused_client()

    def unhide(self):
        self.parent.map()
        self.title_bar.map()
        self.window.map()
        self.make_first()
        self.upper()
        self.wm.give_focus(self)

    def iconify(self):
        if self.states.icccm_state == self.iconified:
            return

        self.states.icccm_state = self.iconified
        self.window.set_wm_state(state=Xutil.IconicState, icon=0)
        self.hide()
        for transient in self.transient_clients:
            if transient.states.icccm_state != self.withdrawn:
                transient.iconify()

    def uniconify(self):
        if self.states.icccm_state != self.iconified:
            return

        self.states.icccm_state = self.normal
        self.window.set_wm_state(state=Xutil.NormalState, icon=0)
        self.unhide()
        for transient in self.transient_clients:
            transient.uniconify()

    def shade(self):
        if self.states.reparented and not self.states.shaded:
            self.parent.configure(height=self.title_bar_height)
            self.geometry = self.parent.get_geometry()
            self.states.shaded = True

    def unshade(self):
        if self.states.reparented and self.states.shaded:
            geo = self.window.get_geometry()
            self.parent.configure(height=self.title_bar_height + geo.height)
            self.geometry = self.parent.get_geometry()
            self.states.shaded = False

    def configure_shape(self):
        # always true, but we never know...
        if self.states.reparented:
            self.window.shape_select_input(1)
            rectangles = self.window.shape_get_rectangles(
                shape.ShapeBounding).rectangles
            if len(rectangles) > 1:
                # a shaped window has no border
                self.parent_border_width = 0
                self.parent.configure(border_width=self.parent_border_width)
                # show the correct window shape
                self.title_bar.shape_select_input(1)
                self.title_bar.shape_rectangles(
                    shape.ShapeSet, shape.ShapeBounding, 1,
                    0, 0,
                    [(0, 0,
                      self.wm.screen.width_in_pixels, self.title_bar_height)])
                self.update_shape()

    def update_shape(self):
        # no need to update anything
        # if there is neither a parent nor a title bar
        if self.states.reparented:
            self.parent.shape_combine(shape.ShapeSet, shape.ShapeBounding,
                                      self.title_bar, shape.ShapeBounding,
                                      0, 0)
            self.parent.shape_combine(shape.ShapeUnion, shape.ShapeBounding,
                                      self.window, shape.ShapeBounding,
                                      0, self.title_bar_height)

    def get_mini_icon(self, size=None):
        if size is None:
            size = (Client.title_bar_height, Client.title_bar_height)
        wm_icon = self.window.get_full_property(atomsdb._NET_WM_ICON, 0)
        if wm_icon:
            value = wm_icon.value
            width = value[0]
            height = value[1]
            data = value[2:width*height+2].tostring()
            data_rgb = ''
            for i in xrange(0, len(data), 4):
                data_rgb += data[i + 2] + data[i + 1] + data[i]
            image = Image.fromstring('RGB', (width, height),
                                     data_rgb).resize(size, Image.ANTIALIAS)
            return image
        elif (self.wm_hints is not None and
              self.wm_hints.flags & Xutil.IconPixmapHint):
            try:
                geo = self.wm_hints.icon_pixmap.get_geometry()
            except BadDrawable:
                return None
            data = self.wm_hints.icon_pixmap.get_image(0, 0,
                                                       geo.width, geo.height,
                                                       ZPixmap, 4294967295).data
            if len(data) != geo.width * geo.height * 4:
                return None
            data_rgb = ''
            for i in xrange(0, len(data), 4):
                data_rgb += data[i + 2] + data[i + 1] + data[i]
            image = Image.fromstring('RGB', (geo.width, geo.height),
                                     data_rgb).resize(size, Image.ANTIALIAS)
            return image
        return None
