# -*- coding: utf-8 -*-

#  PycaWM
#  debug.py
#  Copyright (c) 2007 Vincent Rasneur, Anaël Verrier

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; version 3 only.

#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import sys

if sys.version_info < (2, 5):
    raise ImportError('debug module needs at least Python 2.5!')

import inspect

from time import strftime
from types import StringTypes

class Debug(object):
    """Smart (and simple) debugging and logging system.
    When the log method is called, retrieve useful information
    (file, line, object, method, ...) from the caller's python frames.

    This Debug object is a singleton.
    """
    def __new__(cls, *args, **kwargs):
        if '_inst' not in vars(cls):
            inst = super(Debug, cls).__new__(cls, *args, **kwargs)
            inst.stream = None
            inst.active = False
            inst.time = True
            inst.timefmt = "%H:%M:%S"
            cls._inst = inst
        return cls._inst

    def __init__(self, **kwargs):
        if len(kwargs) > 0:
            self.configure(**kwargs)

    def configure(self, filename=None, filemode='a', active=True,
                  time=True, timefmt=None):
        # filename can be a filename string, a (c)StringIO or a file object
        if type(filename) in StringTypes:
            self.stream = open(filename, filemode)
        elif hasattr(filename, 'write'):
            self.stream = filename
        else:
            self.stream = sys.stderr
        self.active = active
        self.time = time
        if timefmt:
            self.timefmt = timefmt

    def __del__(self):
        # no stream opened, no need to clean it...
        if self.stream is None or self.stream.closed:
            return
        # reload sys if not already here
        import sys
        # do not forget to close the stream properly
        if not hasattr(self.stream, 'fileno') or \
                self.stream.fileno() not in (io.fileno() for io
                                             in (sys.stdout, sys.stderr)):
            # stringio or real file?
            try:
                # try to close it in some way...
                self.stream.flush()
                self.stream.close()
            except AttributeError:
                pass

    def enable(self):
        self.active = True

    def disable(self):
        self.active = False

    def _logback(self, frame):
        co = frame.f_code
        fun = co.co_name

        self.stream.write('File: "%s", line %s, ' %
                          (inspect.getsourcefile(frame), frame.f_lineno))

        try:
            try:
                # get the 1st arg
                first = inspect.getargs(co)[0][0]
                # retrieve the object from the arg name
                obj = frame.f_locals[first]
                # get the method which called the log function
                meth = getattr(obj, fun)
            except (IndexError, AttributeError):
                # the log function was not called by an object method
                self.stream.write('in %s:\n' % fun)
            else:
                # is this the real method from the caller?
                if co == meth.im_func.func_code:
                    self.stream.write('at object %s, in method %s:\n' %
                                      (obj, fun))
                else:
                    # 1st arg was an object with a method named the same
                    # as the function which called the log function (tricky!)
                    self.stream.write('in %s:\n' % fun)
        except (IndexError, KeyError):
            # real error... but we handle it anyway...
            self.stream.write('no caller? :\n')

    def log(self, msg, recursive=True, skipframe=0):
        if self.stream is None or not self.active:
            return

        time = ''
        if self.time:
            time = ' at ' + strftime(self.timefmt)

        self.stream.write('Debug message' + time + ':\n')

        # get the caller frame
        frame = inspect.currentframe().f_back

        while skipframe != 0:
            frame = frame.f_back
            skipframe -= 1

        i = 0
        if recursive:
            frames = list()
            while frame is not None:
                frames.insert(0, frame)
                frame = frame.f_back
            for frame in frames:
                self.stream.write(' ' + i * '|' + '-> ')
                self._logback(frame)
                i += 1
        elif frame is not None:
            self._logback(frame)
            i += 1

        self.stream.write(i * ' ' + '=> Debug: %s\n' % msg)

# definition for some shortcuts

def configure(**kwargs):
    Debug().configure(**kwargs)

def enable():
    Debug().enable()

def disable():
    Debug().disable()

def log(msg, recursive=True, skipframe=1):
    Debug().log(msg, recursive, skipframe)
