# -*- coding: utf-8 -*-

#  PycaWM
#  movepos.py
#  Copyright (c) 2007 Vincent Rasneur, Anaël Verrier

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; version 3 only.

#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import sys

if sys.version_info < (2, 5):
    raise ImportError('movepos module needs at least Python 2.5!')

from functools import partial
from Xlib import X, display

from pycawm.hookmanager import (add_pre_hook, add_post_hook,
                                add_instance_post_hook,
                                remove_pre_hook, remove_post_hook,
                                remove_instance_post_hook)
from pycawm.wrappers import DisplayWrapper
from pycawm import Client

class MovePos(object):
    def __init__(self, wm):
        self.wm = wm
        self.display = display.Display(self.wm.display.get_display_name())
        self.display_wrapper = DisplayWrapper(self.display, self.handle_events)
        self.wm.displays.append(self.display_wrapper)
        self.screen = self.display.screen()
        self.root = self.screen.root
        self.font = self.display.open_font('fixed')
        self.window = self.root.create_window(
            0, 0,
            50, 20,
            2, self.screen.root_depth,
            event_mask=X.ExposureMask,
            background_pixel=self.wm.get_color('DimGray'),
            override_redirect=True)
        self.gc = self.window.create_gc(foreground=self.screen.white_pixel,
                                        font=self.font)
        self.moved_client = None
        add_post_hook(wm.start_moving, self.start_moving)
        add_pre_hook(wm.button_release_reply, self.end_moving)
        self.draw_window()
        self.display.sync()

    def destroy(self):
        remove_post_hook(self.wm.start_moving, self.start_moving)
        remove_pre_hook(self.wm.button_release_reply, self.end_moving)
        if self.moved_client is not None:
            self.end_moving(force=True)
        self.window.kill_client()
        self.font.close()
        self.gc.free()

    def start_moving(self, wm, client):
        self.window.map()
        self.window.configure(stack_mode=X.Above)
        self.moved_client = client
        add_instance_post_hook(client.move, self.draw_window)

    def end_moving(self, wm=None, event=None, force=False):
        if force or self.wm.state == self.wm.moving:
            self.window.unmap()
            remove_instance_post_hook(self.moved_client.move, self.draw_window)
            self.moved_client = None

    def draw_window(self, *args, **kwargs):
        self.window.clear_area()
        if self.moved_client is not None:
            self.window.image_text(
                self.gc, 5, 15, ''.join([str(self.moved_client.geometry.x), 'x',
                                         str(self.moved_client.geometry.y)]))

    def handle_events(self):
        while self.display.pending_events():
            event = self.display.next_event()
            if event.type == X.Expose:
                self.draw_window()
